from logging import getLogger
from typing import Generic, TypeVar
from .Cache import Cache


T = TypeVar("T")


logger = getLogger(__name__)


class Repository(Generic[T]):
    def __init__(self, klass: T) -> None:
        """Create a repository for the specified `klass`."""
        logger.debug(f"Constructing the Repository[{klass}]")
        self._klass: T = klass
        self.cache: Cache = Cache()

    def cache_key(self, *args, **kwargs) -> str:
        """Returns the key to which the instance was cached. The `*args` and `**kwargs` arguments should be the same as the `klass` constructor's arguments."""
        msg = "make_cache_key method was not overridden"
        logger.error(msg)
        raise NotImplementedError(msg)

    def create(self, *args, **kwargs) -> T:
        """When saving the instance to the cache, its key is created by the overriden `make_cache_key` method."""
        logger.debug("Making the cache key")
        cache_key: str = self.cache_key(*args, **kwargs)

        logger.debug(f"Checking the cache for key '{cache_key}'")
        if self.cache.has_key(cache_key):
            return self.get_one(cache_key)

        logger.debug(f"Constructing a new {self._klass}")
        instance: T = self._klass(*args, **kwargs)

        logger.debug(f"Caching the new {self._klass} instance to key '{cache_key}'")
        self.cache.save(cache_key, instance)
        return self.get_one(cache_key)

    def get_all(self) -> list[T]:
        logger.debug(f"Getting all instances of {self._klass}")
        return self.cache.get_all()

    def get_one(self, cache_key: str) -> T:
        logger.debug(f"Getting the {self._klass} cache entry for key '{cache_key}'")
        return self.cache.get_one(str(cache_key))



from logging import getLogger
from typing import Generic, TypeVar


T = TypeVar("T")


logger = getLogger(__name__)


class Cache(Generic[T]):
    def __init__(self) -> None:
        logger.debug(f"Constructing the Cache")  # TODO add type to log
        self._cache: dict[str, T] = {}

    def clear_all(self) -> None:
        """Clears the cache. Returns True."""
        logger.debug("Clearing the Cache")
        self._cache = {}

    def clear_one(self, key: str) -> None:
        """Removes one value corresponding to the key in the cache."""
        logger.debug(f"Clearing '{key}' from Cache")
        key = str(key)
        self._cache[key] = None

    def get_all(self) -> list[T]:
        """Returns all values in the cache."""
        logger.debug(f"Getting all Cache entries")
        return [self._cache[key] for key in self._cache]

    def get_one(self, key: str) -> T:
        """Returns the value cache by `key`."""
        key = str(key)
        logger.debug(f"Getting the '{key}' Cache entry")
        if not self.has_key(key):
            msg = f"'{key}' is not cached"
            logger.error(msg)
            raise KeyError(msg)
        return self._cache[key]

    def has_key(self, key: str) -> bool:
        """Returns True if the `key` is already in use. Otherwise, returns False."""
        return bool(str(key) in self._cache)

    def save(self, key: str, value: T) -> None:
        """Save one key/value pair to the cache."""
        logger.debug(f"Caching to the '{key}' Cache")
        self._cache[str(key)] = value


class Cacheable:
    def __init__(self, *args: str, **kwargs: str) -> None:
        self.cache = Cache()



from logging import getLogger
from spacy.tokens import Doc
from ..Config import Config
from .Match import Match, RawMatch


PATTERN_SET_ALL_MATCHES: str = "all"
PATTERN_SET_ONE_MATCH: str = "one"
PATTERN_SET_LONGEST_MATCH: str = "longest"


logger = getLogger(__name__)


class MatchSet:
    def __init__(
        self,
        raw_matches: list[RawMatch],
        doc: Doc,
        best_match: str = "",
        how_many_matches: str = "",
    ) -> None:
        logger.debug(f"Constructing the MatchSet for '{doc.text}'")
        self.matches: list[Match] = [Match(raw_match, doc) for raw_match in raw_matches]
        self.doc: Doc = doc
        self.best_match: str = best_match
        self.how_many_matches: str = how_many_matches

    def __repr__(self):
        return f"<MatchSet ({len(self.all)}): {[repr(m) for m in self.all]}"

    @property
    def result(self) -> list[Match]:
        if not self.matches:
            return []

        if self.how_many_matches == PATTERN_SET_ONE_MATCH:
            return [self.best]
        elif self.how_many_matches == PATTERN_SET_ALL_MATCHES:
            return self.all
        else:
            msg = f"Invalid value for 'how_many_matches': {self.how_many_matches}"
            logger.error(msg)
            raise ValueError(msg)

    @property
    def all(self) -> list[Match]:
        logger.debug(f"Getting all {len(self.matches)} matches")
        return self.matches

    @property
    def best(self) -> Match:
        logger.info(
            f"Getting the best match of these {len(self.matches)}: {self.matches}"
        )
        longest_match_setting = PATTERN_SET_LONGEST_MATCH

        if self.best_match.upper() == longest_match_setting.upper():
            return self.longest
        else:
            msg = f"Invalid value for 'best_match': {self.best_match}"
            logger.error(msg)
            raise ValueError(msg)

    @property
    def longest(self) -> Match:
        logger.info(
            f"Getting the longest match of these {len(self.matches)}: {self.matches}"
        )
        distances: list[int] = [match.end - match.start for match in self.matches]
        largest_distance_i: int = distances.index(max(distances))
        longest_match: Match = self.matches[largest_distance_i]
        logger.info(f"Got the longest match: {longest_match}")
        return longest_match



from logging import getLogger
from typing import Optional
from unittest import TestCase
from .Detector import Detector
from ..patterns import PatternSet


logger = getLogger(__name__)


class DetectorTester(TestCase):
    def _should_skip_pset(self, pattern_set: PatternSet) -> tuple[bool, str]:
        """Determine if the patternset's tests should be skipped. This operates at the patternset level."""
        should_skip: Optional[bool] = None
        skip_reason: Optional[str] = None

        try:
            skip_setting: str = pattern_set.skip_tests
            should_skip = bool(len(skip_setting))
            skip_reason = skip_setting
        except KeyError:
            should_skip = False
            skip_reason = ""
        except Exception as e:
            raise e

        return (should_skip, skip_reason)

    def _should_skip_test(self, test) -> tuple[bool, str]:
        """Determine if the patternset's tests should be skipped. This operates at the test level."""
        should_skip: Optional[bool] = None
        skip_reason: Optional[str] = None

        try:
            skip_setting: str = test["skip"]
            should_skip = bool(len(skip_setting))
            skip_reason = skip_setting
        except KeyError:
            should_skip = False
            skip_reason = ""
        except Exception as e:
            raise e

        return (should_skip, skip_reason)

    def _get_input(self, test) -> str:
        """Extract the input from the test entry."""
        return test["input"]

    def _get_expected_results(self, test, pattern_set: PatternSet) -> tuple[list[str], list[str]]:
        """Extract the expected rulenames and text spans from the test entry."""
        expected_rulenames = test["rulenames"] if "rulenames" in test else []
        expected_spans = test["spans"] if "spans" in test else []

        # Validate the presence of at least one key containing expected results
        if not expected_rulenames and not expected_spans:
            valid_keys = ["rulenames", "spans"]
            err_msg = f"The test must have at least one of these keys: {valid_keys}"
            raise KeyError(err_msg)

        # TODO refine and re-enable this
        # Validate the length of expected results in relation to the expected number of matches
        # expects_many_rulenames = bool(len(expected_rulenames) > 1)
        # expects_many_spans = bool(len(expected_spans) > 1)
        # if expects_many_rulenames or expects_many_spans:
        #     how_many_matches = pattern_set.meta["how_many_matches"]
        #     one_match_setting = Config().prop_str("PATTERN_SET_ONE_MATCH")

        #     if how_many_matches.upper() == one_match_setting.upper():
        #         err_msg = "The patternset expects only one match, but the test expects multiple matches"
        #         raise ValueError(err_msg)
        return (expected_rulenames, expected_spans)

    def _run_test(self, test, detector: Detector) -> None:
        """The main entrypoint to run an individual test."""
        input: str = self._get_input(test)

        # Determine if the test should be skipped
        (should_skip, skip_reason) = self._should_skip_test(test)
        if should_skip:
            msg = f"Skipping patternset '{detector.pattern_set.name}': '{input}'"
            if bool(len(skip_reason)):
                msg = msg + f" ({skip_reason})"
                print(msg)
            return

        # Extract expected results, run matcher, and compare to actual results
        # Separate each test using subTest to prevent failures from stopping other tests
        with self.subTest(f"{detector.pattern_set.name}:{input}"):
            (expected_rulenames, expected_spans) = self._get_expected_results(test, detector.pattern_set)
            rulenames: list[str] = []
            spans: list[str] = []

            for match in detector(input):
                rulenames.append(match.rulename)
                spans.append(str(match.span))

            if expected_rulenames:
                self.assertListEqual(rulenames, expected_rulenames)
            if expected_spans:
                self.assertListEqual(spans, expected_spans)

    def run_tests(self, detector: Detector) -> Optional[tuple[str, int, int]]:
        """The main entrypoint to run all of the tests in a patternset."""
        pattern_set: PatternSet = detector.pattern_set
        (should_skip, skip_reason) = self._should_skip_pset(pattern_set)
        if should_skip:
            msg = f"Skipping patternset '{pattern_set.name}'"
            if bool(len(skip_reason)):
                msg = msg + f": {skip_reason}"
                print(msg)
            return None

        # Separate using subTest to prevent failures from stopping the run
        num_pass = 0
        num_fail = 0
        if pattern_set.tests:
            with self.subTest(pattern_set.name):
                for test in pattern_set.tests:
                    try:
                        self._run_test(test, detector)
                        num_pass += 1
                    except Exception as e:
                        input: str = self._get_input(test)
                        print("======================================================================")
                        print(f"FAIL: [{pattern_set.name}: {input}]")
                        print("----------------------------------------------------------------------")
                        print("ACTUAL vs EXPECTED\n")
                        print(e)
                        print("----------------------------------------------------------------------\n")
                        num_fail += 1
        return (pattern_set.name, num_pass, num_fail)



# The default language model for spaCy's nlp
DATASET: str = "en_core_web_lg"


# The default format for INFO-level logs
LOGGER_FORMAT_BASIC: str = "[%(asctime)s] %(message)s"


# The default format for DEBUG-, WARN-, and ERROR-level logs
LOGGER_FORMAT_DETAILED: str = "[%(asctime)s][%(module)s:%(funcName)s:%(lineno)d][%(levelname)s] %(message)s"


# The default log level
LOGGER_DEFAULT_LEVEL: int = 30


# The default patternset file extension type
PATTERN_SET_FILE_EXTENSION: str = "yaml"


# The default dirpath for the internal patternset files
PATTERN_SET_HOST_DIR: str = "patternsets"


# The default for patternset meta's best_match when undefined
PATTERN_SET_DEFAULT_BEST_MATCH: str = "longest"


# The default for patternset meta's how_many_matches when undefined
PATTERN_SET_DEFAULT_HOW_MANY_MATCHES: str = "all"



from logging import getLogger
from typing import Any, TextIO
from yaml import FullLoader as FullYamlLoader, load as load_yaml
from ..utils import singleton
from .FileLoader import FileLoader


logger = getLogger(__name__)


@singleton
class YamlLoader(FileLoader):
    def load(self, file: TextIO) -> Any:
        logger.debug(f"Loading the YAML file: {file}")
        return load_yaml(file, Loader=FullYamlLoader)



from logging import getLogger
from spacy.tokens import Doc
from .Nlp import Nlp


logger = getLogger(__name__)


class Input:
    """A helper class for tokenizing text and fragmenting the result into a `list[Doc]`. All properties must return a `list[Doc]` for compatibility purposes, hence the `docs` property."""

    def __init__(self, raw_input: str, extract_noun_chunks: bool = False) -> None:
        logger.debug(f"Constructing the Input for '{raw_input}'")
        self.raw: str = raw_input
        self.doc: Doc = Nlp()(raw_input)
        self.extract_noun_chunks: bool = extract_noun_chunks

    @property
    def docs(self) -> list[Doc]:
        """Wrap the core Doc in a list for compatibility purposes."""
        logger.debug(f"Returning the single Doc in a list[Doc]")
        return [self.doc]

    @property
    def noun_chunks(self) -> list[Doc]:
        """Get all of the nouns as Docs from the core Doc."""
        logger.debug(f"Returning the noun chunk list[Span] as a list[Doc]")
        return [span.as_doc() for span in self.doc.noun_chunks]

    @property
    def fragments(self) -> list[Doc]:
        """Automatically select the most appropriate fragmenting attribute."""
        logger.debug("Fragmenting the Input")
        if self.extract_noun_chunks:
            return self.noun_chunks
        return self.docs



from logging import getLogger
from spacy import load as spacy_load
from spacy.language import Language
from spacy.tokens import Doc, Span
from typing import Union
from .defaults import DATASET
from .utils import singleton


logger = getLogger(__name__)


@singleton
class Nlp:
    def __init__(self, dataset_label: str = DATASET):
        self.label: str = dataset_label
        self._nlp: Language = spacy_load(self.label)

    def __call__(self, *args, **kwargs):
        return self._nlp(*args, **kwargs)


def get_doc(text: Union[str, Doc, Span]) -> Doc:
    """Coerce a `str`, `Doc`, or `Span` into a `Doc`. Raises a `TypeError` when `text` is none of these classes."""
    logger.info(f"Getting the Doc for '{text}' (text_type)")
    if isinstance(text, Doc):
        logger.debug("Skipping tokenizing")
        return text
    elif isinstance(text, Span):
        logger.debug("Converting the Span to a Doc")
        return text.as_doc()
    elif isinstance(text, str):
        logger.debug("Tokenizing")
        return Nlp()(text)
    else:
        msg = f"Cannot get the Doc for '{text}' ({type(text)})"
        logger.error(msg)
        raise TypeError(msg)



from logging import getLogger
from typing import Any, TextIO


logger = getLogger(__name__)


# @singleton  # the parent inherits the child's singleton decorator
class FileLoader:
    def __call__(self, filepath: str) -> dict[str, Any]:
        with open(filepath, "r") as file:
            return self.load(file)

    def load(self, file: TextIO) -> dict[str, Any]:
        msg = f"load(file) was not implemented"
        logger.error(msg)
        raise NotImplementedError(msg)



from functools import cached_property
from logging import getLogger
from spacy import explain
from spacy.tokens import Doc, Span
from ..Nlp import Nlp


MatchId = int
Start = int
End = int
RawMatch = tuple[MatchId, Start, End]
SpanFeatures = dict[str, str]


logger = getLogger(__name__)


class Match:
    """A helper class that wraps the output of the spaCy Matcher and provides useful attributes."""

    def __init__(self, raw_match: RawMatch, doc: Doc) -> None:
        logger.debug(f"Constructing the Match for '{doc.text}'")
        (match_id, start, end) = raw_match
        self.match_id: MatchId = match_id
        self.start: Start = start
        self.end: End = end
        self.doc: Doc = doc

    def __repr__(self):
        return f"<{self.rulename}: {self.span}>"

    @cached_property
    def rulename(self) -> str:
        return Nlp()._nlp.vocab.strings[self.match_id]

    @cached_property
    def span(self) -> Span:
        return self.doc[self.start : self.end]

    @cached_property
    def span_features(self) -> SpanFeatures:
        """Extract useful attributes from the matching `Span`."""
        logger.debug(f"Parsing the '{self.span}' Span")
        return {
            "span": self.span,
            "phrase": self.span.text,
            "root": self.span.root.text,
            "root_head": self.span.root.head.text,
            "pos": self.span.root.pos_,
            "tag": self.span.root.tag_,
            "dep": self.span.root.dep_,
            "phrase_lemma": self.span.lemma_,
            "root_lemma": self.span.root.lemma_,
            "pos_desc": explain(self.span.root.pos_),  # type: ignore # explain is untyped
            "tag_desc": explain(self.span.root.tag_),  # type: ignore # explain is untyped
            "dep_desc": explain(self.span.root.dep_),  # type: ignore # explain is untyped
        }



from logging import getLogger
from spacy.tokens import Token
from typing import Union


PatternData = dict[str, Union[str, list[Token]]]


logger = getLogger(__name__)


class Pattern:
    def __init__(self, data: PatternData):
        self.rulename: str = data["rulename"]
        logger.debug(f"Constructing the '{self.rulename}' Pattern")
        self.tokens: list[Token] = data["tokens"]



from logging import getLogger
from spacy.matcher import Matcher as SpacyMatcher
from spacy.tokens import Doc
from ..matches import RawMatch
from ..Nlp import Nlp


logger = getLogger(__name__)


class RawMatcher:
    """A callable Matcher for getting all Matches for the given Doc."""

    def __init__(self) -> None:
        logger.debug("Constructing the RawMatcher")
        self._matcher = SpacyMatcher(Nlp()._nlp.vocab, validate=True)

    def __call__(self, doc: Doc) -> list[RawMatch]:
        """Get all matches for the given `Doc`."""
        raw_matches: list[RawMatch] = self._matcher(doc)
        logger.info(f"Found {len(raw_matches)} raw matches: {raw_matches}")
        return raw_matches



from logging import getLogger
from spacy import explain
from spacy.language import Language
from tabulate import tabulate


logger = getLogger(__name__)


def token_table(
    nlp: Language,
    sentence: str,
    pos: bool = True,
    tag: bool = True,
    dependency: bool = True,
    lemma: bool = True,
) -> str:
    """Print the linguistics features of each word in a sentence.
    If pos is True, then print the part-of-speech (POS). Defaults to True.
    If tag is True, then print the tag. Defaults to True.
    If dependency is True, then print the dependencices. Defaults to True.
    If lemma is True, then print the lemma. Defaults to True.
    """
    logger.debug(f"Creating the token table for `{sentence}`")

    logger.debug("Creating the table headers")
    headers = []
    headers.append("Word")
    if pos:
        headers.append("POS")
        headers.append("POS Definition")
    if tag:
        headers.append("Tag")
        headers.append("Tag Definition")
    if dependency:
        headers.append("Dep.")
        headers.append("Dep. Definition")
    if lemma:
        headers.append("Lemma.")

    logger.debug("Tokenizing")
    tagged_words = nlp(sentence)

    logger.debug("Extracting the features")
    data = []
    for word in tagged_words:
        entry = []
        entry.append(word.text)
        if pos:
            entry.append(word.pos_)
            entry.append(explain(word.pos_))  # type: ignore # explain is untyped
        if tag:
            entry.append(word.tag_)
            entry.append(explain(word.tag_))  # type: ignore # explain is untyped
        if dependency:
            entry.append(word.dep_)
            entry.append(explain(word.dep_))  # type: ignore # explain is untyped
        if lemma:
            entry.append(word.lemma_)
        data.append(entry)
    logger.debug("Returning the token table")
    return tabulate(data, headers=headers, tablefmt="github")



from logging import getLogger
from os import listdir, path
from typing import Union
from .Config import Config
from .defaults import DATASET, LOGGER_DEFAULT_LEVEL
from .detectors import Detector, DetectorRepository, DetectorTester
from .logger import configure_logger
from .matches import Match
from .Nlp import Nlp
from .utils import Filepath, token_table


class GrammarDetector:
    """This class is the entrypoint for loading in patternset files and evaluating text input. It contains the `DetectorRepository` under the hood, which in turn contains the `Detectors`. By running `GrammarDetector.__call__(self, input: str)`, the text input will be compared against both the provided `patternsets` (via the `patternset_path` keyword argument) and the built-in `patternsets`. Extracting the `Detectors` from the `GrammarDetector` is possible via the `detectors: list[Detector]` property but unnecessary.

    Each resulting `Match` has two properties:
    1) `rulename`   -- (str) The name of the `Pattern` that best matches the input
    2) `span`       -- (str) The exact text that triggered the `Pattern`
    """

    def __init__(
        self,
        builtins: bool = True,
        dataset: str = DATASET,  # en_core_web_lg
        patternset_path: str = "",
        verbose: bool = False,
        very_verbose: bool = False,
    ) -> None:
        """Create an instance of the `GrammarDetector`.

        Keyword arguments:
        builtins            -- (bool) If True, include built-in patternsets (default True)
        dataset             -- (str) The spaCy dataset used to create the global `nlp: Language` (default 'en_core_web_lg')
        patternset_path     -- (str) A filepath or dirpath string pointing to a patternset or collection of patternsets (default '')
        verbose             -- (bool) If True, log INFO-level messages; `very_verbose` takes priority over `verbose` (default False)
        very_verbose        -- (bool) If True, log DEBUG-level messages; `very_verbose` takes priority over `verbose` (default False)
        """
        self._is_configured = False
        self._is_loaded = False

        self.builtins: bool = builtins
        self.dataset: str = dataset
        self.patternset_path: str = patternset_path
        self.verbose: bool = verbose
        self.very_verbose: bool = very_verbose

        self.config = Config()
        self.detector_repo = DetectorRepository()

        self._configure()
        self._load()

    def __call__(self, input: str) -> dict[str, Union[str, list[Match]]]:
        """Returns a dict of `Match`es after running all `Detectors` on the input string. One of the two ways to evaluate text for grammatical features. Use this `GrammarDetector.__call__()` method to evaluate text. 

        Arguments:
        input -- (str) The sentence or chunk of text to be analyzed
        """
        matches: dict[str, Union[str, list[Match]]] = {}
        matches["input"] = input
        for detector in self.detectors:
            matches[detector.name] = detector(input)
        return matches

    @property
    def detectors(self) -> list[Detector]:
        """Returns all `Detectors`. One of the two ways to evaluate text for grammatical features. Use the `Detector.__call__` method to evaluate text."""
        if not self._is_loaded:
            raise RuntimeError(f"_load() must be called before accessing this property")
        return self.detector_repo.get_all()

    def _configure(self) -> None:
        """Configure the logger. Must be called before calling `_load()`."""
        log_level: int = LOGGER_DEFAULT_LEVEL
        if self.very_verbose:
            log_level = 10
        elif self.verbose:
            log_level = 20
        configure_logger(log_level)

        self.logger = getLogger(__name__)
        self._is_configured = True

    def _load(self) -> None:
        """Load the NLP and `PatternSets`. If the constructor's `builtins` is True, then the internal `PatternSets` provided with this class will be loaded. If the constructor's `patternset_path` is a valid filepath/dirpath string, then those `PatternSets` will also be loaded."""
        if not self._is_configured:
            raise RuntimeError(f"_configure() must be called before calling _load()")

        # NLP
        self.nlp = Nlp(self.dataset)

        # Patternsets
        patternset_filepaths: list[str] = []

        ## Internal patternsets
        if self.builtins:
            for fpath in Config().internal_patternset_filepaths:
                fp = Filepath(fpath)
                patternset_filepaths.append(fp.filepath)

        ## External patternsets
        if self.patternset_path:
            if path.isdir(self.patternset_path):
                for fpath in listdir(self.patternset_path):
                    fp = Filepath(f"{self.patternset_path}/{fpath}")
                    patternset_filepaths.append(fp.filepath)
            elif path.isfile(self.patternset_path):
                patternset_filepaths.append(self.patternset_path)
            else:
                msg = f"patternset_path expects a directory or file but got: '{self.patternset_path}"
                self.logger.error(msg)
                raise ValueError(msg)

        # Detectors
        [self.detector_repo.create(fpath) for fpath in patternset_filepaths]
        self._is_loaded = True

    def run_tests(self, internal_tests: bool = False) -> None:
        tester = DetectorTester()
        results = []

        for detector in self.detectors:
            # Internal patternsets
            feature_names = [Filepath(fn).filename for fn in self.config.internal_patternset_filenames]
            if detector.name in feature_names:
                if internal_tests:
                    results.append(tester.run_tests(detector))

            # External patternsets
            else:
                results.append(tester.run_tests(detector))
        
        if results:
            for (name, num_pass, num_fail) in results:
                print(f"{name} = PASS: {num_pass}, FAIL: {num_fail}")
        else:
            print("No tests found")

    def token_table(self, input: str) -> str:
        return token_table(self.nlp._nlp, input)



from functools import reduce
from operator import concat


# Source: https://stackoverflow.com/a/45323085
def flatten(a):
    """Flatten a multi-dimensional List to one-dimensional List."""
    return reduce(concat, a)



from logging import getLogger
from pprint import pformat
from sys import argv
from time import time
from .GrammarDetector import GrammarDetector
from .utils import token_table


def main() -> None:
    start_time: float = time()
    sentences: list[str] = []

    # Validate the input
    if len(argv) > 1:
        sentences = argv[1:]
    else:
        raise ValueError("No sentences were provided")

    # Create the detectors
    grammar_detector = GrammarDetector()
    logger = getLogger(__name__)

    # Run the detectors
    count: int = 0
    features = {}
    for sentence in sentences:
        logger.info(f"Sentence {count}: '{sentence}'")
        sentence_start_time: float = time()
        features = grammar_detector(sentence)
        print(grammar_detector.token_table(sentence))
        sentence_finish_time: float = time()
        count += 1

    finish_time: float = time()
    logger.info(f"Total run time: {finish_time - start_time:.2f}s")

    logger.info("Detected features:")
    print(features)


if __name__ == "__main__":
    main()
